# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import psycopg2

from .. import backend

__all__ = ["Backend"]


class Backend(backend.Backend):
    def connect(self):
        if self.name == "":
            raise AssertionError("You must specify a value for database NAME in settings file.")
        if self.user == "":
            raise AssertionError("You must specify a value for database USER in settings file.")
        conn_string = ["dbname=postgres"]
        if self.user:
            conn_string.append("user=%s" % self.user)
        if self.password:
            conn_string.append("password='%s'" % self.password)
        if self.host:
            conn_string.append("host=%s" % self.host)
        if self.port:
            conn_string.append("port=%s" % self.port)

        self.connection = psycopg2.connect(" ".join(conn_string))
        try:
            self.connection.autocommit = True
        except Exception:
            self.connection.set_isolation_level(0)
        self.connection.set_client_encoding("UTF8")
        self.cursor = self.connection.cursor()

    def close(self):
        if self.cursor:
            self.cursor.close()
        if self.connection:
            self.connection.close()
        self.connection = None
        self.cursor = None

    def createdb(self):
        self.cursor.execute("CREATE DATABASE %s OWNER %s ENCODING 'UTF8'" % (self.name, self.user))

    def dropdb(self):
        self.cursor.execute("DROP DATABASE IF EXISTS %s" % (self.name,))



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import os
from importlib import import_module


class ChdirMixin:
    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument("module", nargs="?", default=None, help="path/module to process")

    def handle(self, *args, **options):
        module = options.get("module")
        if module is not None:
            if os.path.exists(module):
                path = module
            else:
                path = os.path.dirname(import_module(module).__file__)
            path = os.path.abspath(path)
            os.chdir(path)

        return super().handle(**options)



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

# Copyright (C) 2007 Michael Trier
# ForeignKeySearchInput taken from django-extensions
import datetime
import re

from django import forms
from django.contrib.admin.widgets import (
    AdminDateWidget as DateWidget,
    AdminSplitDateTime as DateTimeWidget,
    AdminTimeWidget as TimeWidget,
)
from django.forms.utils import flatatt
from django.utils.dates import MONTHS
from django.utils.encoding import smart_str
from django.utils.html import escape
from django.utils.safestring import mark_safe

__all__ = [
    "DateWidget",
    "TimeWidget",
    "DateTimeWidget",
    "GroupedSelect",
    "SelectYearWidget",
    "SelectMonthYearWidget",
    "TimeDeltaWidget",
]


RE_DATE = re.compile(r"(\d{4})-(\d\d?)-(\d\d?)$")


# taken and adapted from http://djangosnippets.org/snippets/200/
class GroupedSelect(forms.Select):
    def render(self, name, value, attrs=None, choices=()):
        if value is None:
            value = ""
        final_attrs = self.build_attrs(attrs, name=name)
        output = ["<select%s>" % flatatt(final_attrs)]
        str_value = smart_str(value)
        for group_label, group in self.choices:
            if group_label:  # should belong to an optgroup
                group_label = smart_str(group_label)
                output.append('<optgroup label="%s">' % escape(group_label))
            for k, v in group:
                option_value = smart_str(k)
                option_label = smart_str(v)
                selected_html = (option_value == str_value) and ' selected="selected"' or ""
                output.append(
                    '<option value="%s"%s>%s</option>' % (escape(option_value), selected_html, escape(option_label)),
                )
            if group_label:
                output.append("</optgroup>")
        output.append("</select>")
        return mark_safe("\n".join(output))


class SelectYearWidget(forms.Widget):
    none_value = (0, "---")
    year_field = "%s_year"

    def __init__(self, attrs=None, years=None, required=True):
        # years is an optional list/tuple of years to use in the "year" select box.
        self.attrs = attrs or {}
        self.required = required
        if years:
            self.years = years
        else:
            this_year = datetime.date.today().year
            self.years = range(this_year, this_year + 10)

    def render(self, name, value, attrs=None):
        try:
            year_val = value.year
        except AttributeError:
            year_val = month_val = None
            if isinstance(value, str):
                match = RE_DATE.match(value)
                if match:
                    year_val, month_val, day_val = [int(v) for v in match.groups()]

        output = []

        if "id" in self.attrs:
            id_ = self.attrs["id"]
        else:
            id_ = "id_%s" % name

        year_choices = [(i, i) for i in self.years]
        if not (self.required and value):
            year_choices.insert(0, self.none_value)
        local_attrs = self.build_attrs(id=self.year_field % id_)
        s = forms.Select(choices=year_choices)
        select_html = s.render(self.year_field % name, year_val, local_attrs)
        output.append(select_html)

        return mark_safe("\n".join(output))

    def id_for_label(self, id_):
        return "%s_year" % id_

    id_for_label = classmethod(id_for_label)

    def value_from_datadict(self, data, files, name):
        y = data.get(self.year_field % name)
        if y == "0":
            return None
        if y:
            return "%s-%s-%s" % (y, 1, 1)
        return data.get(name, None)


class SelectMonthYearWidget(forms.Widget):
    none_value = (0, "---")
    month_field = "%s_month"
    year_field = "%s_year"

    def __init__(self, attrs=None, years=None, required=True):
        # years is an optional list/tuple of years to use in the "year" select box.
        self.attrs = attrs or {}
        self.required = required
        if years:
            self.years = years
        else:
            this_year = datetime.date.today().year
            self.years = range(this_year, this_year + 10)

    def render(self, name, value, attrs=None):
        try:
            year_val, month_val = value.year, value.month
        except AttributeError:
            year_val = month_val = None
            if isinstance(value, str):
                match = RE_DATE.match(value)
                if match:
                    year_val, month_val, day_val = [int(v) for v in match.groups()]

        output = []

        if "id" in self.attrs:
            id_ = self.attrs["id"]
        else:
            id_ = "id_%s" % name

        month_choices = MONTHS.items()
        if not (self.required and value):
            month_choices.append(self.none_value)
        month_choices.sort()
        local_attrs = self.build_attrs(id=self.month_field % id_)
        s = forms.Select(choices=month_choices)
        select_html = s.render(self.month_field % name, month_val, local_attrs)
        output.append(select_html)

        year_choices = [(i, i) for i in self.years]
        if not (self.required and value):
            year_choices.insert(0, self.none_value)
        local_attrs["id"] = self.year_field % id_
        s = forms.Select(choices=year_choices)
        select_html = s.render(self.year_field % name, year_val, local_attrs)
        output.append(select_html)

        return mark_safe("\n".join(output))

    def id_for_label(self, id_):
        return "%s_month" % id_

    id_for_label = classmethod(id_for_label)

    def value_from_datadict(self, data, files, name):
        y = data.get(self.year_field % name)
        m = data.get(self.month_field % name)
        if y == m == "0":
            return None
        if y and m == "0":
            return "%s-%s-%s" % (y, 1, 1)
        if y and m:
            return "%s-%s-%s" % (y, m, 1)
        return data.get(name, None)


class TimeDeltaWidget(forms.MultiWidget):
    """Input accurate timing. IntegerFields for hours, minutes, seconds and milliseconds."""

    def __init__(self, milliseconds=False, attrs=None):
        attrs = attrs if attrs is not None else {"size": 2, "maxlength": 2, "style": "width:auto;"}
        self.show_milliseconds = milliseconds
        self.NUM_FIELDS = 4 if milliseconds else 3
        widgets = [forms.TextInput(attrs=attrs)] * self.NUM_FIELDS
        super().__init__(widgets, attrs)

    def format_output(self, widgets):
        widgets.insert(1, " : ")
        widgets.insert(3, " : ")
        if self.show_milliseconds:
            widgets.insert(5, " . ")
        return mark_safe("".join(widgets))

    def decompress(self, value):
        if value:
            return [
                "%02d" % (value / 3600),
                "%02d" % (value % 3600 / 60),
                "%02d" % (value % 3600 % 60),
                "%03d" % (value % 1 * 1000),
            ][: self.NUM_FIELDS]
        return [None] * self.NUM_FIELDS



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from django.core import mail


class EmailMessage(mail.EmailMessage):
    def __init__(
        self,
        subject="",
        body="",
        from_email=None,
        to=None,
        cc=None,
        bcc=None,
        connection=None,
        attachments=None,
        headers=None,
    ):
        if isinstance(to, str):
            to = [to]
        if isinstance(cc, str):
            cc = [cc]
        if isinstance(bcc, str):
            bcc = [bcc]
        self.cc = cc
        super().__init__(subject, body, from_email, to, bcc, connection, attachments, headers)

    def message(self):
        msg = super().message()
        msg["Cc"] = ", ".join(self.cc)
        return msg

    def recipients(self):
        return self.to + self.cc + self.bcc


class EmailMultiAlternatives(mail.EmailMessage):
    multipart_subtype = "alternative"

    def attach_alternative(self, content, mimetype=None):
        self.attach(content=content, mimetype=mimetype)



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand

User = get_user_model()


class Command(BaseCommand):
    help = "Load a default admin user data: username=admin password=admin"

    def handle(self, *args, **options):
        """Load a default admin user"""
        try:
            admin = User.objects.get(username="admin")
        except User.DoesNotExist:
            admin = User(
                username="admin",
                first_name="admin",
                last_name="admin",
                email="admin@localhost.localdomain",
                is_staff=True,
                is_active=True,
                is_superuser=True,
            )
        admin.set_password("admin")
        admin.save()



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from operator import add, mul

from django import forms
from django.core.exceptions import ValidationError
from django.utils.encoding import smart_str
from django.utils.translation import gettext, gettext_lazy as _

from .widgets import GroupedSelect, TimeDeltaWidget

__all__ = [
    "OrderField",
    "StringField",
    "TextField",
    "GroupedChoiceField",
    "TimeDeltaField",
]


try:
    # new from django-admin-ui branch
    from django.forms import OrderField
except ImportError:

    class OrderField(forms.IntegerField):
        pass


class StringField(forms.CharField):
    pass


class TextField(forms.CharField):
    widget = forms.Textarea


# taken and adapted from http://djangosnippets.org/snippets/200/
class GroupedChoiceField(forms.ChoiceField):
    widget = GroupedSelect

    def __init__(
        self,
        choices=(),
        required=True,
        widget=None,
        label=None,
        initial=None,
        help_text=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            # choices=choices,
            required=required,
            widget=widget,
            label=label,
            initial=initial,
            help_text=help_text,
            *args,
            **kwargs,
        )
        self.choices = choices

    def clean(self, value):
        """
        Validates that the input is in self.choices.
        """
        value = super().clean(value)
        if value in (None, ""):
            value = ""
        value = smart_str(value)
        if value == "":
            return value
        valid_values = []
        for group_label, group in self.choices:
            valid_values += [str(k) for k, v in group]
        if value not in valid_values:
            raise ValidationError(gettext("Select a valid choice. That choice is not one of the available choices."))
        return value


class TimeDeltaField(forms.MultiValueField):
    """Input accurate timing. Interface with models.TimeDeltaField."""

    LABELS = [_("Hours"), _("Minutes"), _("Seconds"), _("Milliseconds")]
    SECONDS = [60 * 60, 60, 1, 0.001]

    def __init__(self, milliseconds=True, *args, **kwargs):
        if not milliseconds:
            self.LABELS = self.LABELS[:3]
            self.SECONDS = self.SECONDS[:3]
        self.widget = TimeDeltaWidget(milliseconds=milliseconds)
        fields = [forms.CharField(label=label) for label in self.LABELS]
        super().__init__(fields, *args, **kwargs)

    def compress(self, value):
        from functools import reduce

        if value:
            return "{}".format(reduce(add, map(lambda x: mul(*x), zip(map(float, value), self.SECONDS))))
        return None



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

# taken and adapted from django-cms
# Copyright (c) 2008, Batiste Bieler
"this is the locale selecting middleware that will look at accept headers"

import re

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect
from django.utils import translation
from django.utils.cache import patch_vary_headers
from django.utils.deprecation import MiddlewareMixin

from fluo.settings import NO_LOCALE_PATTERNS

SUB = re.compile(
    r'<a([^>]+)href="/(?!(%s|%s|%s))([^"]*)"([^>]*)>'
    % ("|".join(map(lambda l: l[0] + "/", settings.LANGUAGES)), settings.MEDIA_URL[1:], settings.STATIC_URL[1:]),
)
SUB2 = re.compile(
    r'<form([^>]+)action="/(?!(%s|%s|%s))([^"]*)"([^>]*)>'
    % ("|".join(map(lambda l: l[0] + "/", settings.LANGUAGES)), settings.MEDIA_URL[1:], settings.STATIC_URL[1:]),
)
SUPPORTED = dict(settings.LANGUAGES)
START_SUB = re.compile(r"^/(%s)/(.*)" % "|".join(map(lambda l: l[0], settings.LANGUAGES)))
NO_LOCALE_SUB = re.compile(r"^(%s|%s)(.*)" % ("|".join(NO_LOCALE_PATTERNS), settings.STATIC_URL))
LANGUAGE_COOKIE_NAME = settings.LANGUAGE_COOKIE_NAME


def has_lang_prefix(path):
    check = START_SUB.match(path)
    if check is not None:
        return check.group(1)
    else:
        return False


def skip_translation(path):
    check = NO_LOCALE_SUB.match(path)
    if check is not None:
        return check.group(1)
    else:
        return False


def get_default_language(language_code=None):
    """
    Returns default language depending on settings.LANGUAGE_CODE merged with
    best match from settings.LANGUAGES

    Returns: language_code

    Raises ImproperlyConfigured if no match found
    """

    if not language_code:
        language_code = settings.LANGUAGE_CODE

    languages = dict(settings.LANGUAGES).keys()

    # first try if there is an exact language
    if language_code in languages:
        return language_code

    # otherwise split the language code if possible, so iso3
    language_code = language_code.split("-")[0]

    if language_code not in languages:
        raise ImproperlyConfigured("No match in LANGUAGES for LANGUAGE_CODE %s" % settings.LANGUAGE_CODE)

    return language_code


def get_language_from_request(request):
    language = request.GET.get("language", request.POST.get("language", None))

    if language:
        if language not in dict(settings.LANGUAGES).keys():
            language = None

    if language is None:
        language = getattr(request, "LANGUAGE_CODE", None)

    if language:
        if language not in dict(settings.LANGUAGES).keys():
            language = None

    if language is None:
        language = get_default_language()

    return language


class LocaleMiddleware(MiddlewareMixin):
    def get_language_from_request(self, request):
        changed = False
        prefix = has_lang_prefix(request.path_info)
        if prefix:
            request.path = "/" + "/".join(request.path.split("/")[2:])
            request.path_info = "/" + "/".join(request.path_info.split("/")[2:])
            t = prefix
            if t in SUPPORTED:
                lang = t
                if hasattr(request, "session"):
                    request.session["django_language"] = lang
                else:
                    request.set_cookie(LANGUAGE_COOKIE_NAME, lang)
                changed = True
        else:
            lang = translation.get_language_from_request(request)
        if not changed:
            if hasattr(request, "session"):
                lang = request.session.get("django_language", None)
                if lang in SUPPORTED and lang is not None:
                    return lang
            elif LANGUAGE_COOKIE_NAME in request.COOKIES.keys():
                lang = request.COOKIES.get(LANGUAGE_COOKIE_NAME, None)
                if lang in SUPPORTED and lang is not None:
                    return lang
            if not lang:
                lang = translation.get_language_from_request(request)
        lang = get_default_language(lang)
        return lang

    def process_request(self, request):
        path = str(request.path)
        if skip_translation(path):
            return

        prefix = has_lang_prefix(request.path_info)
        if not prefix:
            return HttpResponseRedirect("/%s%s" % (settings.LANGUAGE_CODE[:2], request.get_full_path()))
        language = self.get_language_from_request(request)
        translation.activate(language)
        request.LANGUAGE_CODE = translation.get_language()

    def process_response(self, request, response):
        path = str(request.path)
        if skip_translation(path):
            return response

        patch_vary_headers(response, ["Accept-Language"])
        translation.deactivate()

        if (
            not skip_translation(path)
            and response.status_code == 200
            and response._headers["content-type"][1].split(";")[0] == "text/html"
        ):
            response.content = SUB.sub(
                r'<a\1href="/%s/\3"\4>' % request.LANGUAGE_CODE,
                response.content.decode("utf-8"),
            )
            response.content = SUB2.sub(
                r'<form\1action="/%s/\3"\4>' % request.LANGUAGE_CODE,
                response.content.decode("utf-8"),
            )
        if response.status_code == 301 or response.status_code == 302:
            if "Content-Language" not in response:
                response["Content-Language"] = translation.get_language()
            location = response._headers["location"]
            prefix = has_lang_prefix(location[1])
            if not prefix and location[1].startswith("/") and not skip_translation(location[1]):
                response._headers["location"] = (
                    location[0],
                    "/%s%s" % (request.LANGUAGE_CODE, location[1]),
                )
        return response



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import os
from importlib import import_module

from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext as _

__all__ = [
    "get_backend",
    "Backend",
    "ImproperlyConfigured",
    "ConnectionError",
    "CreateDBError",
    "DropDBError",
]


class ConnectionError(Exception):
    pass


class CreateDBError(Exception):
    pass


class DropDBError(Exception):
    pass


class Backend:
    def __init__(self, options):
        self.host = options.get("host")
        self.port = options.get("port")
        self.password = options.get("password")
        self.name = options.get("name")
        self.user = options.get("user")
        self.engine = options.get("user")
        self.options = options

    def do_createdb(self):
        raise ImproperlyConfigured(
            "%s.%s.do_createdb is not implemented." % (self.__class__.__module__, self.__class__.__name__),
        )

    def do_dropdb(self):
        raise ImproperlyConfigured(
            "%s.%s.do_dropdb is not implemented." % (self.__class__.__module__, self.__class__.__name__),
        )

    def connect(self):
        raise ImproperlyConfigured(
            "%s.%s.connect is not implemented." % (self.__class__.__module__, self.__class__.__name__),
        )

    def close(self):
        raise ImproperlyConfigured(
            "%s.%s.close is not implemented." % (self.__class__.__module__, self.__class__.__name__),
        )

    def createdb(self):
        raise ImproperlyConfigured(
            "%s.%s.createdb is not implemented." % (self.__class__.__module__, self.__class__.__name__),
        )

    def dropdb(self):
        raise ImproperlyConfigured(
            "%s.%s.dropdb is not implemented." % (self.__class__.__module__, self.__class__.__name__),
        )


class BackendWrapper:
    def __init__(self, database):
        engine = database["ENGINE"].split(".")[-1]
        self.options = {
            "host": database["HOST"],
            "port": database["PORT"],
            "password": database["PASSWORD"],
            "name": database["NAME"],
            "user": database["USER"],
            "engine": engine,
        }
        self.backend = self._get_backend(self.options)

    def connect(self):
        return self.backend.connect()

    def close(self):
        return self.backend.close()

    def createdb(self):
        try:
            print(_("Creating database %(name)s...") % self.options)
            self.backend.createdb()
            print(_("done"))
        except Exception as e:
            raise CreateDBError(_("Cannot create db: %s") % e)

    def dropdb(self):
        try:
            print(_("Dropping database %(name)s...") % self.options)
            self.backend.dropdb()
            print(_("done"))
        except Exception as e:
            raise DropDBError(_("Cannot drop db: %s") % e)

    def _get_backend(self, options):
        engine = options["engine"]
        try:
            module = import_module("fluo.db.backends.%s" % engine)
        except ImportError:
            try:
                module = import_module(".db_commands", engine)
            except ImportError as e_user:
                # The database backend wasn't found. Display a helpful error message
                # listing all possible (built-in) database backends.
                backend_dir = os.path.join(__path__[0], "backends")  # noqa: F821
                try:
                    available_backends = [
                        f
                        for f in os.listdir(backend_dir)
                        if os.path.isdir(os.path.join(backend_dir, f)) and not f.startswith(".")
                    ]
                except EnvironmentError:
                    available_backends = []
                available_backends.sort()
                if engine not in available_backends:
                    error_msg = "%r isn't an available database backend. Available options are: %s\nError was: %s" % (
                        engine,
                        ", ".join(map(repr, available_backends)),
                        e_user,
                    )
                    raise ImproperlyConfigured(error_msg)
                else:
                    raise  # If there's some other error, this must be an error in Django itself.
        try:
            return module.Backend(options)
        except Exception as e:
            print(e)
            raise Backend.ConnectionError("Cannot connect to database %(name)s" % options)


def get_backend(database):
    return BackendWrapper(database)



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from django.contrib.admin import (  # noqa: F401
    AllValuesFieldListFilter,
    BooleanFieldListFilter,
    ChoicesFieldListFilter,
    DateFieldListFilter,
    FieldListFilter,
    ListFilter,
    RelatedFieldListFilter,
    RelatedOnlyFieldListFilter,
    SimpleListFilter,
    autodiscover,
    register,
    site,
)
from django.contrib.admin.options import HORIZONTAL, VERTICAL  # noqa: F401
from django.contrib.contenttypes.admin import (  # noqa: F401
    GenericInlineModelAdmin,
    GenericStackedInline,
    GenericTabularInline,
)

from .base import CopyObject  # noqa: F401
from .models import (  # noqa: F401
    CategoryModelAdmin,
    ModelAdmin,
    OrderedModelAdmin,
    ReadOnlyInlineMixin,
    ReadOnlyMixin,
    ReadOnlyModelAdmin,
    ReadOnlyStackedInline,
    ReadOnlyTabularInline,
    StackedInline,
    TabularInline,
    TreeOrderedModelAdmin,
)
from .sites import AdminSite  # noqa: F401



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from django.conf import settings
from django.utils.encoding import iri_to_uri

DATABASES = getattr(settings, "DATABASES")
DEBUG = getattr(settings, "DEBUG")

MEDIA_URL = iri_to_uri(getattr(settings, "MEDIA_URL"))
MEDIA_ROOT = getattr(settings, "MEDIA_ROOT")
FLUO_MEDIA_URL = getattr(settings, "FLUO_MEDIA_URL", MEDIA_URL + "fluo/")

STATIC_URL = iri_to_uri(getattr(settings, "STATIC_URL"))
STATIC_ROOT = getattr(settings, "STATIC_ROOT")
FLUO_STATIC_URL = getattr(settings, "FLUO_STATIC_URL", STATIC_URL + "fluo/")

SERVE_MEDIA_FILES = getattr(settings, "SERVE_MEDIA_FILES", True)

LANGUAGES = getattr(settings, "LANGUAGES")
LOGGING_FORMAT = getattr(settings, "LOGGING_FORMAT", "%(asctime)s %(levelname)s %(message)s")
LOGGING_FILENAME = getattr(settings, "LOGGING_FILENAME", "/tmp/fluo.log")

DEFAULT_PERMISSIONS = (
    "list",
    "view",
)
try:
    DEFAULT_PERMISSIONS += list(getattr(settings, "DEFAULT_PERMISSIONS"))
except Exception:
    pass

NO_LOCALE_PATTERNS = [MEDIA_URL]
try:
    NO_LOCALE_PATTERNS += getattr(settings, "NO_LOCALE_PATTERNS")
except Exception:
    pass



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import re
from urllib.parse import quote

from django.contrib.auth.models import AbstractBaseUser, BaseUserManager, PermissionsMixin, UserManager
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core import validators
from django.core.mail import send_mail
from django.db import models
from django.db.models import Max
from django.template.defaultfilters import slugify
from django.utils import timezone
from django.utils.translation import get_language, gettext_lazy as _

from fluo import settings

from . import fields

__all__ = [
    "OrderedModel",
    "TreeOrderedModel",
    "TimestampModel",
    "I18NProxy",
    "I18NModel",
    "TranslationModel",
    "CategoryModelManager",
    "CategoryModel",
    "CategoryTranslationModel",
    "GenericModel",
    "AbstractUser",
    "BaseUserManager",
    "UserManager",
]


class OrderedModel(models.Model):
    ordering = fields.OrderField(
        default=0,
        blank=True,
        db_index=True,
        verbose_name=_("ordering"),
        help_text=_("Ordered"),
    )

    class Meta:
        abstract = True
        ordering = ["-ordering"]

    def save(self, *args, **kwargs):
        if not self.ordering:
            self.ordering = self.get_max_ordering() + 1
        super().save(*args, **kwargs)

    @property
    def brothers_and_me(self):
        return self.__class__._default_manager.all()

    @property
    def brothers(self):
        return self.brothers_and_me.exclude(pk=self.id)

    def get_max_ordering(self):
        ordering = self.brothers.aggregate(max=Max("ordering"))["max"]
        return 0 if ordering is None else ordering


class TreeOrderedModel(OrderedModel):
    parent = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="children",
        verbose_name=_("Parent node"),
        help_text=_("The parent node of this field."),
    )

    class Meta:
        abstract = True

    @property
    def brothers_and_me(self):
        if self.parent:
            return self.__class__._default_manager.filter(parent=self.parent)
        else:
            return self.__class__._default_manager.filter(parent__isnull=True)


class TimestampModel(models.Model):
    created_at = fields.CreationDateTimeField(verbose_name=_("created"))
    last_modified_at = fields.ModificationDateTimeField(verbose_name=_("modified"))

    class Meta:
        abstract = True


class I18NProxy:
    def __init__(self, tr, original):
        self._tr = tr
        self._original = original

    def __getattr__(self, name):
        attr = getattr(self._tr, name, None)
        if not attr or (attr and (attr == "" or attr == "")):
            attr = getattr(self._original, name)
        return attr


class I18NModel(models.Model):
    def translate(self, language=None):
        try:
            language = language or get_language()[:2]
            return I18NProxy(self.translations.get(language__startswith=language), self)
        except (models.ObjectDoesNotExist, AttributeError, TypeError):
            return self

    class Meta:
        abstract = True


class TranslationModel(models.Model):
    language = models.CharField(max_length=5, choices=settings.LANGUAGES, db_index=True, verbose_name=_("language"))

    class Meta:
        abstract = True
        ordering = ["language"]


class CategoryModelQuerySet(models.QuerySet):
    def default(self):
        return self.get(default=True)

    def active(self):
        return self.filter(status=fields.STATUS_ACTIVE)

    def inactive(self):
        return self.filter(status=fields.STATUS_INACTIVE)


class CategoryModelManager(models.Manager.from_queryset(CategoryModelQuerySet)):
    use_for_related_fields = True
    silence_use_for_related_fields_deprecation = True


class CategoryModel(OrderedModel):
    status = fields.StatusField()

    objects = CategoryModelManager()

    name = models.CharField(unique=True, max_length=255)
    slug = models.SlugField(
        unique=True,
        editable=False,
        verbose_name=_("slug"),
        help_text=_(
            'A "slug" is a unique URL-friendly title for the object automatically generated from the "name" field.',
        ),  # noqa: E501
    )
    default = models.BooleanField(default=False, verbose_name=_("default"), help_text=_("Is the default one?"))

    class Meta:
        abstract = True
        base_manager_name = "objects"
        ordering = ["name"]

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        self.slug = slugify(self.name)
        super().save(*args, **kwargs)
        if self.default:
            for c in self.__class__._default_manager.exclude(pk=self.id):
                c.default = False
                c.save(*args, **kwargs)
        try:
            c = self.__class__._default_manager.get(default=True)
        except models.ObjectDoesNotExist:
            self.default = True
            super().save(*args, **kwargs)


class CategoryTranslationModel(TranslationModel):
    name = models.CharField(max_length=255)

    class Meta:
        abstract = True


class GenericModel(models.Model):
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, verbose_name=_("Object Type"))
    object_id = models.PositiveIntegerField(verbose_name=_("Object ID"))
    content_object = GenericForeignKey("content_type", "object_id")

    class Meta:
        abstract = True


class AbstractUser(AbstractBaseUser, PermissionsMixin):
    """
    An abstract base class implementing a fully featured User model with
    admin-compliant permissions.

    Username, password and email are required. Other fields are optional.
    """

    objects = UserManager()

    username = models.CharField(
        max_length=255,
        unique=True,
        validators=[validators.RegexValidator(re.compile(r"^[\w.@+-]+$"), _("Enter a valid username."), "invalid")],
        verbose_name=_("username"),
        help_text=_("Required. 255 characters or fewer. Letters, numbers and @/./+/-/_ characters"),
    )
    first_name = models.CharField(max_length=255, blank=True, verbose_name=_("first name"))
    last_name = models.CharField(max_length=255, blank=True, verbose_name=_("last name"))
    email = models.EmailField(max_length=255, blank=True, verbose_name=_("email address"))
    is_staff = models.BooleanField(
        default=False,
        verbose_name=_("staff status"),
        help_text=_("Designates whether the user can log into this admin site."),
    )
    is_active = models.BooleanField(
        default=True,
        verbose_name=_("active"),
        help_text=_(
            "Designates whether this user should be treated as active. Unselect this instead of deleting accounts.",
        ),  # noqa: E501
    )
    date_joined = models.DateTimeField(default=timezone.now, verbose_name=_("date joined"))

    USERNAME_FIELD = "username"
    REQUIRED_FIELDS = ["email"]

    class Meta:
        abstract = True
        verbose_name = _("user")
        verbose_name_plural = _("users")

    def get_absolute_url(self):
        return "/users/%s/" % quote(self.username)

    def get_full_name(self):
        """
        Returns the first_name plus the last_name, with a space in between.
        """
        full_name = "%s %s" % (self.first_name, self.last_name)
        return full_name.strip()

    def get_short_name(self):
        "Returns the short name for the user."
        return self.first_name

    def email_user(self, subject, message, from_email=None):
        """
        Sends an email to this User.
        """
        send_mail(subject, message, from_email, [self.email])

    def get_display_name(self):
        if self.first_name and self.last_name:
            full_name = "%s %s" % (self.first_name, self.last_name)
            return full_name.strip()
        return self.username



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from django import template
from django.templatetags.static import static
from django.utils.encoding import iri_to_uri
from django.utils.safestring import mark_safe

from fluo.settings import MEDIA_URL

register = template.Library()


class MediaNode(template.Node):
    def __init__(self, filename, args):
        self.filename = filename
        self.args = args

    def render(self, context):
        filename = self.filename.resolve(context)
        args = " %s" % " ".join(self.args) if self.args else ""
        script = filename if filename.startswith(("https://", "http://", "//")) else static(iri_to_uri(filename))
        return mark_safe(self.fmt % {"script": script, "args": args})


def media_tag(parser, token, node_class):
    bits = token.split_contents()
    if len(bits) < 2:
        raise template.TemplateSyntaxError("'%s' takes at least one argument, the name of the script." % bits[0])

    filename = parser.compile_filter(bits[1])
    args = [bit for bit in bits[2:]]

    return node_class(filename, args)


class CssNode(MediaNode):
    fmt = '<link rel="stylesheet" type="text/css" href="%(script)s"%(args)s/>'


@register.tag
def css(parser, token):
    return media_tag(parser, token, CssNode)


class JsNode(MediaNode):
    fmt = '<script src="%(script)s"%(args)s></script>'


@register.tag
def js(parser, token):
    return media_tag(parser, token, JsNode)


@register.simple_tag
def static_url():
    """Returns the string contained in the setting STATIC_URL."""
    return static("")


@register.simple_tag
def media_url():
    """Returns the string contained in the setting MEDIA_URL."""
    return MEDIA_URL



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from django.conf.urls import handler400, handler403, handler404, handler500
from django.urls import (
    NoReverseMatch,
    Resolver404,
    ResolverMatch,
    get_script_prefix,
    resolve,
    reverse as django_reverse,
)
from django.utils.functional import lazy
from django.utils.http import urlencode

try:
    from django.urls import include
except ImportError:
    from django.conf.urls import include

try:
    from django.urls import path, re_path
except ImportError:
    path, re_path = None, None

try:
    from django.urls import re_path as url
except ImportError:
    from django.conf.urls import url

try:
    from django.urls import URLPattern, URLResolver
except ImportError:
    from django.urls import RegexURLPattern as URLPattern, RegexURLResolver as URLResolver


__all__ = [
    "handler400",
    "handler403",
    "handler404",
    "handler500",
    "path",
    "re_path",
    "url",
    "include",
    "NoReverseMatch",
    "URLPattern",
    "URLResolver",
    "ResolverMatch",
    "Resolver404",
    "get_script_prefix",
    "reverse",
    "reverse_lazy",
    "resolve",
    "UrlsMixin",
]


class UrlsMixin:
    def get_urls(self):
        raise NotImplementedError

    @property
    def urls(self):
        return self.get_urls()


def reverse(viewname, *, args=None, kwargs=None, request=None, format=None, data=None, **extra):
    if format is not None:
        kwargs = kwargs or {}
        kwargs["format"] = format
    url = django_reverse(viewname, args=args, kwargs=kwargs, **extra)

    if data is not None:
        url += "?" + urlencode(data)

    if request:
        url = request.build_absolute_uri(url)

    return url


reverse_lazy = lazy(reverse, str)



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType, ContentTypeManager
from django.db.models import *  # noqa: F401,F403
from django.db.models import __all__ as django_all

from .fields import *  # noqa: F401,F403
from .fields import __all__ as fields_all
from .models import *  # noqa: F401,F403
from .models import __all__ as models_all

try:
    from .enums import *  # noqa: F401,F403
    from .enums import __all__ as enums_all
except ImportError:
    enums_all = []

__all__ = [
    *[
        model
        for model in django_all
        if model not in ["Choices", "EmailField", "IntegerChoices", "SlugField", "TextChoices", "URLField"]
    ],
    *enums_all,
    *fields_all,
    *models_all,
    "GenericForeignKey",
    "GenericRelation",
    "ContentType",
    "ContentTypeManager",
]



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from functools import update_wrapper

from django.contrib import admin
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.http import Http404, HttpResponseRedirect
from django.utils.translation import gettext as _

__all__ = [
    "AdminSite",
]


class AdminSite(admin.AdminSite):
    def get_urls(self):
        from django.urls import path

        def wrap(view, cacheable=False):
            def wrapper(*args, **kwargs):
                return self.admin_view(view, cacheable)(*args, **kwargs)

            return update_wrapper(wrapper, view)

        return [
            path("r/<int:content_type_id>/<path:object_id>/", wrap(self.view_on_site), name="view_on_site"),
        ] + super().get_urls()

    def view_on_site(self, request, content_type_id, object_id):
        """
        Redirect to an object's page based on a content-type ID and an object ID.
        """
        # Look up the object, making sure it's got a get_absolute_url() function.
        try:
            content_type = ContentType.objects.get(pk=content_type_id)
            if not content_type.model_class():
                raise Http404(_("Content type %(ct_id)s object has no associated model") % {"ct_id": content_type_id})
            obj = content_type.get_object_for_this_type(pk=object_id)
        except (ObjectDoesNotExist, ValueError):
            raise Http404(
                _("Content type %(ct_id)s object %(obj_id)s doesn't exist")
                % {"ct_id": content_type_id, "obj_id": object_id},
            )

        try:
            get_absolute_url = obj.get_absolute_url
        except AttributeError:
            raise Http404(
                _("%(ct_name)s objects don't have a get_absolute_url() method") % {"ct_name": content_type.name},
            )
        absurl = get_absolute_url()

        return HttpResponseRedirect(absurl)



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from django.shortcuts import get_list_or_404, get_object_or_404, redirect, render as render_to_response
from django.template import Template
from django.template.loader import render_to_string

__all__ = [
    "get_object_or_404",
    "get_list_or_404",
    "redirect",
    "render_to_string",
    "render_to_response",
    "render_from_string",
]


def render_from_string(template_string, context=None, request=None):
    return Template(template_string).render(context, request)



from __future__ import annotations

import operator
from functools import reduce

try:
    from django.contrib.admin.utils import lookup_spawns_duplicates
except ImportError:
    from django.contrib.admin.utils import lookup_needs_distinct as lookup_spawns_duplicates
from django.contrib.admin.views.autocomplete import AutocompleteJsonView
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django.http import Http404, JsonResponse


class RelatedSearchJsonView(AutocompleteJsonView):
    search_fields = []

    def get_queryset(self):
        qs = self.model_admin.get_queryset(self.request)
        qs, search_use_distinct = self.get_search_results(self.request, qs, self.search_fields, self.term)
        if search_use_distinct:
            qs = qs.distinct()
        return qs

    def get_search_results(self, request, queryset, search_fields, search_term):
        """
        Return a tuple containing a queryset to implement the search
        and a boolean indicating if the results may contain duplicates.
        """
        # override ModelAdmin.get_search_results to use a custom search_fields

        # Apply keyword searches.
        def construct_search(field_name):
            if field_name.startswith("^"):
                return "%s__istartswith" % field_name[1:]
            elif field_name.startswith("="):
                return "%s__iexact" % field_name[1:]
            elif field_name.startswith("@"):
                return "%s__search" % field_name[1:]
            # Use field_name if it includes a lookup.
            opts = queryset.model._meta
            lookup_fields = field_name.split(LOOKUP_SEP)
            # Go through the fields, following all relations.
            prev_field = None
            for path_part in lookup_fields:
                if path_part == "pk":
                    path_part = opts.pk.name
                try:
                    field = opts.get_field(path_part)
                except FieldDoesNotExist:
                    # Use valid query lookups.
                    if prev_field and prev_field.get_lookup(path_part):
                        return field_name
                else:
                    prev_field = field
                    if hasattr(field, "get_path_info"):
                        # Update opts to follow the relation.
                        opts = field.get_path_info()[-1].to_opts
            # Otherwise, use the field with icontains.
            return "%s__icontains" % field_name

        use_distinct = False
        if search_fields and search_term:
            orm_lookups = [construct_search(str(search_field)) for search_field in search_fields]
            for bit in search_term.split():
                or_queries = [models.Q(**{orm_lookup: bit}) for orm_lookup in orm_lookups]
                queryset = queryset.filter(reduce(operator.or_, or_queries))
            use_distinct |= any(
                lookup_spawns_duplicates(queryset.model._meta, search_spec) for search_spec in orm_lookups
            )

        return queryset, use_distinct

    def get(self, request, *args, **kwargs):
        """
        Return a JsonResponse with search results of the form:
        {
            results: [{id: "123" text: "foo"}],
            pagination: {more: true}
        }
        """
        if not self.search_fields:
            raise Http404("%s must have search_fields for the related_search_view." % type(self.model_admin).__name__)
        if not self.has_perm(request):
            return JsonResponse({"error": "403 Forbidden"}, status=403)

        self.term = request.GET.get("term", "")
        self.object_list = self.get_queryset()
        context = self.get_context_data()
        return JsonResponse(
            {
                "results": [{"id": str(obj.pk), "text": str(obj)} for obj in context["object_list"]],
                "pagination": {"more": context["page_obj"].has_next()},
            },
        )



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import json

from django.http import *  # noqa: F401,F403
from django.http import __all__ as http_all

from .utils.json import JSONEncoder

__all__ = [
    *http_all,
    "JsonResponse",
]


class JsonResponse(HttpResponse):  # noqa: F405
    def __init__(
        self,
        content=None,
        content_type="application/json; charset=utf-8",
        status=200,
        indent=2,
    ):
        """
        return JsonResponse(content={'status': 200, 'message': '', 'data': [] })
        """
        if content is None:
            content = {}
        data = json.dumps(content, indent=indent, cls=JSONEncoder)
        super().__init__(content=data, content_type=content_type, status=status)
        self["Content-Length"] = len(data)



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

# Original code taken from django-extensions
# Copyright (c) 2007 Michael Trier
import getpass

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError
from django.db.models import Q

User = get_user_model()


class Command(BaseCommand):
    help = "Change a user password."
    requires_model_validation = False

    def add_arguments(self, parser):
        parser.add_argument("username", help="user username or email")
        parser.add_argument("password", nargs="?", default="", help="new password")

    def handle(self, *args, **options):
        username = options.get("username")

        try:
            user = User.objects.get(Q(username__iexact=username) | Q(email__iexact=username))
        except User.DoesNotExist:
            raise CommandError("user %s does not exist" % username)

        print("Changing password for user %s" % user.username)
        password = options.get("password")

        p1 = p2 = password

        while "" in (p1, p2) or p1 != p2:
            p1 = getpass.getpass()
            p2 = getpass.getpass("Password (again): ")
            if p1 != p2:
                print("Passwords do not match, try again")
            elif "" in (p1, p2):
                raise CommandError("aborted")

        user.set_password(p1)
        user.save()

        return "Password changed successfully for user %s\n" % user.username



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import sys

from django.core.management.base import BaseCommand, CommandError
from django.db import DEFAULT_DB_ALIAS, close_old_connections, connections

from fluo.db.backend import get_backend


class DatabaseCommand(BaseCommand):
    output_transaction = True
    requires_system_checks = []

    message = """You have requested to drop "%(name)s" database.
This will IRREVERSIBLY DELETE all data currently in the "%(name)s" database.
Are you sure you want to do this?"""
    ask_message = """

    Type 'yes' to continue, or 'no' to cancel: """
    error_message = """Database %(name)s couldn't be dropped. Possible reasons:
  * The database isn't running or isn't configured correctly.
  * The database is in use by another user.
The full error: %(error)s"""
    default = "yes"
    should_ask = True

    def add_arguments(self, parser):
        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_false",
            dest="interactive",
            default=True,
            help="Tells Django to NOT prompt the user for input of any kind.",
        )
        parser.add_argument(
            "--database",
            action="store",
            dest="database",
            default=DEFAULT_DB_ALIAS,
            help='Nominates a database. Defaults to the "default" database.',
        )

    def handle(self, *args, **options):
        close_old_connections()

        db = options.get("database")
        connection = connections[db]
        interactive = options.get("interactive")
        settings = dict(connection.settings_dict)
        name = settings["NAME"]
        if interactive and self.should_ask:
            msg = "".join([self.message % {"name": name}, self.ask_message])
            confirm = input(msg)
        else:
            confirm = self.default

        if confirm == "yes":
            backend = get_backend(settings)
            backend.connect()
            try:
                self.execute_sql(backend, **options)
            except Exception as e:
                raise CommandError(CommandError(self.error_message % {"name": name, "error": e})).with_traceback(
                    sys.exc_info()[2],
                )
            finally:
                backend.close()

            self.post_execute(**options)

    def execute_sql(self, backend):
        raise NotImplementedError

    def post_execute(self, **options):
        pass



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import django
from django.apps import apps

if apps.get_containing_app_config("django.contrib.postgres"):
    from django.contrib.postgres.fields.array import *  # noqa
    from django.contrib.postgres.fields.array import __all__ as array_all
    from django.contrib.postgres.fields.citext import *  # noqa
    from django.contrib.postgres.fields.citext import __all__ as citext_all
    from django.contrib.postgres.fields.hstore import *  # noqa
    from django.contrib.postgres.fields.hstore import __all__ as hstore_all

    if django.VERSION >= (3, 1):
        jsonb_all = []
    else:
        from django.contrib.postgres.fields.jsonb import *  # noqa
        from django.contrib.postgres.fields.jsonb import __all__ as jsonb_all
    from django.contrib.postgres.fields.ranges import *  # noqa
    from django.contrib.postgres.fields.ranges import __all__ as ranges_all

    postgres__all__ = [
        *array_all,
        *[k for k in citext_all if k not in ["CIEmailField"]],
        *hstore_all,
        *jsonb_all,
        *ranges_all,
    ]
else:
    from django.db import models

    postgres__all__ = ["CIText", "CICharField", "CITextField"]

    class CIText:
        pass

    class CICharField(models.CharField):
        pass

    class CITextField(models.TextField):
        pass


__all__ = postgres__all__



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import subprocess
import sys

from django.conf import settings
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "pip wrapper"
    requires_model_validation = False

    def add_arguments(self, parser):
        parser.add_argument(
            "--install",
            "-i",
            action="store_true",
            default=True,
            dest="install",
            help="intall package",
        )
        parser.add_argument(
            "--upgrade",
            "-u",
            action="store_true",
            default=False,
            dest="upgrade",
            help="upgrade package",
        )
        parser.add_argument(
            "--uninstall",
            "-r",
            action="store_true",
            default=False,
            dest="uninstall",
            help="uninstall package",
        )
        parser.add_argument(
            "--verbose",
            action="store_true",
            default=False,
            dest="verbose",
            help="verbose",
        )
        parser.add_argument("pkgs", metavar="pkgs", nargs="+", help="package(s)")

    def handle(self, *pkgs, **options):
        pkgs = pkgs or options.get("pkgs")

        if options.get("uninstall"):
            self.uninstall(pkgs, options)
        elif options.get("install") or options.get("upgrade"):
            self.install(pkgs, options)

    def install(self, pkgs, options):
        cmd = [
            sys.executable,
            "-m",
            "pip",
            "install",
        ]
        if options.get("upgrade"):
            cmd.append("--upgrade")
        if options.get("verbosity") > 1 or options.get("verbose"):
            cmd.append("--verbose")
        cmd.extend(["--target=%s" % settings.LIB_DIR])
        cmd.extend(pkgs)
        subprocess.call(cmd)

    def uninstall(self, pkgs, options):
        cmd = [
            sys.executable,
            "-m",
            "pip",
            "uninstall",
        ]
        cmd.extend(pkgs)
        subprocess.call(cmd)



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from __future__ import annotations

from django import forms
from django.contrib.admin.widgets import AutocompleteMixin
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

__all__ = [
    "AdminImageFileWidget",
    "RelatedSearchSelect",
    "RelatedSearchSelectMultiple",
]


class AdminImageFileWidget(forms.FileInput):
    def __init__(self, attrs=None):
        super().__init__(attrs)

    def render(self, name, value, attrs=None):
        output = []
        if value and hasattr(value, "url"):
            output.append(
                """
<div>
    <div style="float: left; vertical-align: middle;">
        <a target="_blank" href="%(url)s"><img src="%(url)s" width="40px"/></a>
    </div>
    <div style="padding-left: 20px; float: left; vertical-align: middle;">
        %(text)s <a target="_blank" href="%(url)s">%(value)s</a>
        <br />
        %(change)s """
                % {"text": _("Currently:"), "url": value.url, "value": value, "change": _("Change:")},
            )
        output.append(super().render(name, value, attrs))
        output.append("</div>")
        if value and hasattr(value, "url"):
            output.append("</div>")
        return mark_safe("".join(output))


class RelatedSearchMixin(AutocompleteMixin):
    url_name = "%s:%s_%s_related_search"

    def get_url(self):
        model = self.rel.model
        return reverse(
            self.url_name % (self.admin_site.name, model._meta.app_label, model._meta.model_name),
            args=[self.rel.name],
        )


class RelatedSearchSelect(RelatedSearchMixin, forms.Select):
    pass


class RelatedSearchSelectMultiple(RelatedSearchMixin, forms.SelectMultiple):
    pass



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import django
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.decorators import user_passes_test
from django.http import HttpResponseBadRequest, HttpResponseForbidden
from django.utils.cache import cache
from django.utils.decorators import method_decorator, wraps

__all__ = [
    "ajax_required",
    "ajax_required_m",
    "login_required",
    "login_required_m",
    "throttle",
    "throttle_m",
]


def ajax_required(func):
    # taken from djangosnippets.org
    """
    AJAX request required decorator
    use it in your views:

    @ajax_required
    def my_view(request):
        ....

    """

    def wrap(request, *args, **kwargs):
        if not request.is_ajax():
            return HttpResponseBadRequest
        return func(request, *args, **kwargs)

    wrap.__doc__ = func.__doc__
    wrap.__name__ = func.__name__
    return wrap


ajax_required_m = method_decorator(ajax_required)


def login_required(function=None, required=False, redirect_field_name=REDIRECT_FIELD_NAME):
    """
    Decorator for views that, if required, checks that the user is logged in and redirect
    to the log-in page if necessary.
    """
    if required:
        if django.VERSION < (1, 11):
            actual_decorator = user_passes_test(lambda u: u.is_authenticated(), redirect_field_name=redirect_field_name)
        else:
            actual_decorator = user_passes_test(lambda u: u.is_authenticated, redirect_field_name=redirect_field_name)
        if function:
            return actual_decorator(function)
        return actual_decorator

    # login not required
    def decorator(view_func):
        def _wrapper(request, *args, **kwargs):
            return function(request, *args, **kwargs)

        return wraps(function)(_wrapper)

    return method_decorator(decorator)


login_required_m = method_decorator(login_required)


def throttle(func, limit=3, duration=900, methods=("POST", "GET", "PUT", "DELETE", "OPTIONS")):
    def inner(request, *args, **kwargs):
        if request.method in methods:
            remote_addr = request.META.get("HTTP_X_FORWARDED_FOR") or request.META.get("REMOTE_ADDR")
            if cache.get(remote_addr) == limit:
                return HttpResponseForbidden("Try slowing down a little.")
            elif not cache.get(remote_addr):
                cache.set(remote_addr, 1, duration)
            else:
                cache.incr(remote_addr)
        return func(request, *args, **kwargs)

    return inner


throttle_m = method_decorator(throttle)



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from django.core.management.base import BaseCommand

User = get_user_model()


class Command(BaseCommand):
    help = "Set some data useful for local development."

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--username",
            "-u",
            action="store",
            dest="username",
            default="admin",
            help="Use username as admin user.",
        )
        parser.add_argument(
            "--password",
            "-p",
            action="store",
            dest="password",
            default="admin",
            help='New password for "admin" user.',
        )
        parser.add_argument(
            "--site",
            "-s",
            action="store",
            dest="site",
            default="1",
            help="Update this SITE_ID domain/name.",
        )
        parser.add_argument(
            "--domain",
            "-d",
            action="store",
            dest="domain",
            default="localhost:8000",
            help="Use this domain/name for SITE_ID.",
        )

    def handle(self, **options):
        user = User.objects.get(username=options.get("username"))
        user.set_password(options.get("password"))
        user.save()

        site = Site.objects.get(pk=options.get("site"))
        domain = options.get("domain")
        site.domain = domain
        site.name = domain
        site.save()



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import sys

from django.http import Http404
from django.utils.deprecation import MiddlewareMixin
from django.views.debug import technical_404_response, technical_500_response


class UserBasedExceptionMiddleware(MiddlewareMixin):
    def process_exception(self, request, exception):
        if request.user.is_superuser:
            if isinstance(exception, Http404):
                return technical_404_response(request, exception)
            return technical_500_response(request, *sys.exc_info())



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import copy

from django.db import transaction
from django.utils.translation import gettext_lazy as _

__all__ = [
    "CopyObject",
]


class CopyObject:
    short_description = _("Duplicate as new")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__name__ = self.__class__.__name__

    @transaction.atomic
    def __call__(self, modeladmin, request, queryset):
        for original in queryset:
            instance = copy.copy(original)
            instance.id = None
            self.update(request, instance, original)
            instance.save()
            self.update_m2m(request, instance, original)
            instance.save()

            translations = getattr(original, "translations", None)
            if translations:
                for translation in translations.all():
                    tr = copy.copy(translation)
                    tr.id = None
                    tr.parent = instance
                    self.update_translation(request, tr, translation)
                    tr.save()

    def update(self, request, instance, original):
        pass

    def update_m2m(self, request, instance, original):
        pass

    def update_translation(self, request, instance, original):
        pass



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import json

from django.core.serializers.json import DjangoJSONEncoder
from django.db.models.query import QuerySet

__all__ = [
    "decoder",
    "encoder",
    "scanner",
    "dump",
    "dumps",
    "load",
    "loads",
]


class JSONEncoder(DjangoJSONEncoder):
    def default(self, obj):  # noqa
        if isinstance(obj, QuerySet):
            return tuple(obj)
        elif hasattr(obj, "tolist"):
            # Numpy arrays and array scalars.
            return obj.tolist()
        elif hasattr(obj, "__getitem__"):
            try:
                return dict(obj)
            except Exception:
                pass
        elif hasattr(obj, "__iter__"):
            return tuple(item for item in obj)
        return super().default(obj)


decoder = json.decoder
encoder = json.encoder
scanner = json.scanner


def dump(
    obj,
    fp,
    skipkeys=False,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    cls=None,
    indent=None,
    separators=None,
    encoding="utf-8",
    default=None,
    **kw,
):
    """Serialize ``obj`` as a JSON formatted stream to ``fp`` (a
    ``.write()``-supporting file-like object).

    If ``skipkeys`` is true then ``dict`` keys that are not basic types
    (``str``, ``unicode``, ``int``, ``long``, ``float``, ``bool``, ``None``)
    will be skipped instead of raising a ``TypeError``.

    If ``ensure_ascii`` is false, then the some chunks written to ``fp``
    may be ``unicode`` instances, subject to normal Python ``str`` to
    ``unicode`` coercion rules. Unless ``fp.write()`` explicitly
    understands ``unicode`` (as in ``codecs.getwriter()``) this is likely
    to cause an error.

    If ``check_circular`` is false, then the circular reference check
    for container types will be skipped and a circular reference will
    result in an ``OverflowError`` (or worse).

    If ``allow_nan`` is false, then it will be a ``ValueError`` to
    serialize out of range ``float`` values (``nan``, ``inf``, ``-inf``)
    in strict compliance of the JSON specification, instead of using the
    JavaScript equivalents (``NaN``, ``Infinity``, ``-Infinity``).

    If ``indent`` is a non-negative integer, then JSON array elements and
    object members will be pretty-printed with that indent level. An indent
    level of 0 will only insert newlines. ``None`` is the most compact
    representation.

    If ``separators`` is an ``(item_separator, dict_separator)`` tuple
    then it will be used instead of the default ``(', ', ': ')`` separators.
    ``(',', ':')`` is the most compact JSON representation.

    ``encoding`` is the character encoding for str instances, default is UTF-8.

    ``default(obj)`` is a function that should return a serializable version
    of obj or raise TypeError. The default simply raises TypeError.

    To use a custom ``JSONEncoder`` subclass (e.g. one that overrides the
    ``.default()`` method to serialize additional types), specify it with
    the ``cls`` kwarg; otherwise ``DjangoJSONEncoder`` is used.

    """
    cls = cls or JSONEncoder
    kwargs = {
        "obj": obj,
        "fp": fp,
        "skipkeys": skipkeys,
        "ensure_ascii": ensure_ascii,
        "check_circular": check_circular,
        "allow_nan": allow_nan,
        "cls": cls,
        "indent": indent,
        "separators": separators,
        "default": default,
    }
    kwargs.update(kw)

    return json.dump(**kwargs)


def dumps(
    obj,
    skipkeys=False,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    cls=None,
    indent=None,
    separators=None,
    encoding="utf-8",
    default=None,
    **kw,
):
    """Serialize ``obj`` to a JSON formatted ``str``.

    If ``skipkeys`` is false then ``dict`` keys that are not basic types
    (``str``, ``unicode``, ``int``, ``long``, ``float``, ``bool``, ``None``)
    will be skipped instead of raising a ``TypeError``.

    If ``ensure_ascii`` is false, then the return value will be a
    ``unicode`` instance subject to normal Python ``str`` to ``unicode``
    coercion rules instead of being escaped to an ASCII ``str``.

    If ``check_circular`` is false, then the circular reference check
    for container types will be skipped and a circular reference will
    result in an ``OverflowError`` (or worse).

    If ``allow_nan`` is false, then it will be a ``ValueError`` to
    serialize out of range ``float`` values (``nan``, ``inf``, ``-inf``) in
    strict compliance of the JSON specification, instead of using the
    JavaScript equivalents (``NaN``, ``Infinity``, ``-Infinity``).

    If ``indent`` is a non-negative integer, then JSON array elements and
    object members will be pretty-printed with that indent level. An indent
    level of 0 will only insert newlines. ``None`` is the most compact
    representation.

    If ``separators`` is an ``(item_separator, dict_separator)`` tuple
    then it will be used instead of the default ``(', ', ': ')`` separators.
    ``(',', ':')`` is the most compact JSON representation.

    ``encoding`` is the character encoding for str instances, default is UTF-8.

    ``default(obj)`` is a function that should return a serializable version
    of obj or raise TypeError. The default simply raises TypeError.

    To use a custom ``JSONEncoder`` subclass (e.g. one that overrides the
    ``.default()`` method to serialize additional types), specify it with
    the ``cls`` kwarg; otherwise ``DjangoJSONEncoder`` is used.

    """
    cls = cls or JSONEncoder
    kwargs = {
        "obj": obj,
        "skipkeys": skipkeys,
        "ensure_ascii": ensure_ascii,
        "check_circular": check_circular,
        "allow_nan": allow_nan,
        "cls": cls,
        "indent": indent,
        "separators": separators,
        "default": default,
    }
    kwargs.update(kw)

    return json.dumps(**kwargs)


def load(
    fp,
    encoding=None,
    cls=None,
    object_hook=None,
    parse_float=None,
    parse_int=None,
    parse_constant=None,
    object_pairs_hook=None,
    **kw,
):
    kwargs = {
        "fp": fp,
        "cls": cls,
        "object_hook": object_hook,
        "parse_float": parse_float,
        "parse_int": parse_int,
        "parse_constant": parse_constant,
        "object_pairs_hook": object_pairs_hook,
    }
    kwargs.update(kw)

    return json.load(**kwargs)


load.__doc__ = json.load.__doc__


def loads(
    s,
    encoding=None,
    cls=None,
    object_hook=None,
    parse_float=None,
    parse_int=None,
    parse_constant=None,
    object_pairs_hook=None,
    **kw,
):
    kwargs = {
        "s": s,
        "cls": cls,
        "object_hook": object_hook,
        "parse_float": parse_float,
        "parse_int": parse_int,
        "parse_constant": parse_constant,
        "object_pairs_hook": object_pairs_hook,
    }
    kwargs.update(kw)

    return json.loads(**kwargs)


loads.__doc__ = json.loads.__doc__



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from ..database import DatabaseCommand


class Command(DatabaseCommand):
    help = "Try to (re)create the database in an empty state!"
    message = """You have requested to create "%(name)s" database."""
    error_message = """Database %(name)s couldn't be created. Possible reasons:
  * The database isn't running or isn't configured correctly.
  * The database alread exists.
The full error: %(error)s"""
    should_ask = False

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--force",
            action="store_true",
            dest="force",
            default=False,
            help="Drop database if exists before (re)create it.",
        )

    def execute_sql(self, backend, **options):
        if options.get("force"):
            backend.dropdb()
        backend.createdb()



from __future__ import annotations

from django.core.exceptions import ValidationError
from django.db.models import enums

__all__ = ["Choices", "ChoicesMeta", "IntChoices", "StrChoices", "StrChoicesMeta"]


class ChoicesMeta(enums.ChoicesMeta):
    def __new__(metacls, classname, bases, classdict):
        # allow implicit value declaration
        new_values = {}
        for name in classdict._member_names:
            value = classdict[name]
            if not isinstance(value, (tuple, list)):
                new_values[name] = (name, value)
        classdict.update(new_values)
        return super().__new__(metacls, classname, bases, classdict)

    @property
    def display_names(cls):
        empty = [{"value": None, "display_name": cls.__empty__}] if hasattr(cls, "__empty__") else []
        return empty + [member.display_name for member in cls]

    @property
    def as_dict(cls):
        return dict(cls.choices)

    # Django-filter checks if choice is a callable to get the choices
    def __call__(cls, *args, **kwargs):
        # If called without arguments (it has no meaning for standard Enum, so we can add meaning):
        if not args and not kwargs:
            return cls.choices
        return super().__call__(*args, **kwargs)


class Choices(enums.Choices, metaclass=ChoicesMeta):
    """Class for creating enumerated choices."""

    @property
    def choice(self):
        return (self.value, self.label)

    @property
    def as_dict(self):
        return {self.value: self.label}

    @property
    def display_name(self):
        return {"value": self.value, "display_name": self.label}

    @classmethod
    def validate(cls, value):
        try:
            cls(value)
        except ValueError as exc:
            raise ValidationError(f"Invalid choice: {cls.__name__}.{value}") from exc


class IntChoices(int, Choices):
    """Integer based choices."""

    pass


class StrChoicesMeta(ChoicesMeta):
    @property
    def max_length(cls):
        return max([len(value) for value in cls.values])


class StrChoices(str, Choices, metaclass=StrChoicesMeta):
    """String based choices."""

    pass



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

# Original code for taken and adapted from
# - django-extensions
#   - CreationDateTimeField
#   - ModificationDateTimeField
from django.core import checks, validators
from django.db import connection, models
from django.db.models import lookups
from django.utils import timezone
from django.utils.encoding import smart_str
from django.utils.translation import gettext_lazy as _

from fluo import forms

from .postgresql import *  # noqa: F401,F403
from .postgresql import __all__ as postgresql_all

__all__ = [
    *postgresql_all,
    "StatusField",
    "STATUS_ACTIVE",
    "STATUS_INACTIVE",
    "STATUS_CHOICES",
    "CreationDateTimeField",
    "ModificationDateTimeField",
    "OrderField",
    "TimeDeltaField",
    "SlugField",
    "StringField",
    "EmailField",
    "URLField",
    "CISlugField",
    "CIStringField",
    "CIEmailField",
    "CIURLField",
]


STATUS_ACTIVE = "active"
STATUS_INACTIVE = "inactive"
STATUS_CHOICES = (
    (STATUS_ACTIVE, _("Active")),
    (STATUS_INACTIVE, _("Inactive")),
)


class CreationDateTimeField(models.DateTimeField):
    """
    By default, sets editable=False, blank=True, default=datetime.now
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("editable", False)
        kwargs.setdefault("blank", True)
        kwargs.setdefault("default", timezone.now)
        super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return "DateTimeField"


class ModificationDateTimeField(CreationDateTimeField):
    """
    By default, sets editable=False, blank=True, default=datetime.now

    Sets value to datetime.now() on each save of the model.
    """

    def pre_save(self, model, add):
        value = timezone.now()
        setattr(model, self.attname, value)
        return value

    def get_internal_type(self):
        return "DateTimeField"


class OrderField(models.IntegerField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("default", 0)
        models.Field.__init__(self, *args, **kwargs)

    def formfield(self, **kwargs):
        defaults = {"form_class": forms.OrderField}
        defaults.update(kwargs)
        return super().formfield(**defaults)


class TimeDeltaField(models.DecimalField):
    description = _("TimeDelta field")

    def __init__(self, milliseconds=False, *args, **kwargs):
        self.milliseconds = milliseconds
        kwargs.setdefault("default", 0)
        kwargs.setdefault("decimal_places", 3)
        kwargs.setdefault("max_digits", 12)
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        defaults = {
            "form_class": forms.TimeDeltaField,
            "milliseconds": self.milliseconds,
        }
        defaults.update(kwargs)
        # skip DecimalField.formfield
        # which injects decimal_places and max_digits
        return models.Field.formfield(self, **defaults)


class StringField(models.Field):
    description = _("String")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(self.max_length, int) and self.max_length > 0:
            self.validators.append(validators.MaxLengthValidator(self.max_length))

    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_max_length_attribute(**kwargs),
        ]

    def formfield(self, **kwargs):
        defaults = {"max_length": self.max_length, "form_class": forms.StringField}
        # see django CharField comment
        if self.null and not connection.features.interprets_empty_strings_as_nulls:
            defaults["empty_value"] = None
        defaults.update(kwargs)
        return super().formfield(**defaults)

    def get_internal_type(self):
        return "CharField" if self.max_length else "TextField"

    def to_python(self, value):
        if isinstance(value, str) or value is None:
            return value
        return smart_str(value)

    def get_prep_value(self, value):
        value = super().get_prep_value(value)
        return self.to_python(value)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["max_length"] = self.max_length
        return name, path, args, kwargs

    def _check_max_length_attribute(self, **kwargs):
        is_integer = isinstance(self.max_length, int)
        is_negative = is_integer and self.max_length <= 0
        is_none = self.max_length is None
        if (is_integer and is_negative) or (not is_integer and not is_none):
            errors = [
                checks.Error(
                    "StringField must define a 'max_length' attribute (can be None, 0 or a positive integer).",
                    hint=None,
                    obj=self,
                    id="fluo.E1",
                ),
            ]
        else:
            errors = []
        return errors


class URLField(StringField):
    default_validators = [validators.URLValidator()]
    description = _("URL")

    def formfield(self, **kwargs):
        # As with StringField, this will cause URL validation to be performed twice.
        defaults = {
            "form_class": forms.URLField,
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)


class SlugField(StringField):
    default_validators = [validators.validate_slug]
    description = _("Slug")

    def __init__(self, *args, db_index=True, allow_unicode=False, **kwargs):
        self.allow_unicode = allow_unicode
        if self.allow_unicode:
            self.default_validators = [validators.validate_unicode_slug]
        super().__init__(*args, db_index=db_index, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(SlugField, self).deconstruct()
        if self.db_index is False:
            kwargs["db_index"] = False
        else:
            del kwargs["db_index"]
        if self.allow_unicode is not False:
            kwargs["allow_unicode"] = self.allow_unicode
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        defaults = {"form_class": forms.SlugField, "allow_unicode": self.allow_unicode}
        defaults.update(kwargs)
        return super().formfield(**defaults)


class EmailField(StringField):
    default_validators = [validators.validate_email]
    description = _("Email address")

    def formfield(self, **kwargs):
        defaults = {
            "form_class": forms.EmailField,
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)


class StatusField(StringField):
    def __init__(
        self,
        choices=STATUS_CHOICES,
        max_length=None,
        default=STATUS_ACTIVE,
        blank=False,
        null=False,
        verbose_name=_("status"),
        help_text=_("Is active?"),
    ):
        super().__init__(
            choices=choices,
            max_length=max_length,
            default=default,
            blank=blank,
            null=null,
            verbose_name=verbose_name,
            help_text=help_text,
        )


class CIStringField(CIText, StringField):  # noqa: F405
    pass


class CIURLField(CIText, URLField):  # noqa: F405
    pass


class CISlugField(CIText, SlugField):  # noqa: F405
    pass


class CIEmailField(CIText, EmailField):  # noqa: F405
    pass


for field in [CIStringField, CIURLField, CISlugField, CIEmailField]:
    field.register_lookup(lookups.IExact, lookup_name="exact")
    field.register_lookup(lookups.IContains, lookup_name="contains")
    field.register_lookup(lookups.IStartsWith, lookup_name="startswith")
    field.register_lookup(lookups.IEndsWith, lookup_name="endswith")
    field.register_lookup(lookups.IRegex, lookup_name="regex")



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Creates permissions for all installed apps that need permissions.
"""

from __future__ import annotations

import logging

from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS
from django.db.models import signals

from .. import settings

__all__ = ["DatabaseCommand"]


log = logging.getLogger("fluo")


class DatabaseCommand(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--noinput",
            action="store_false",
            dest="interactive",
            default=True,
            help="Tells Django to NOT prompt the user for input of any kind.",
        )
        parser.add_argument(
            "--no-utf8",
            action="store_true",
            dest="no_utf8_support",
            default=False,
            help="Tells Django to not create a UTF-8 charset database",
        )
        parser.add_argument(
            "-U",
            "--user",
            action="store",
            dest="user",
            default=None,
            help="Use another user for the database then defined in settings.py",
        )
        parser.add_argument(
            "-P",
            "--password",
            action="store",
            dest="password",
            default=None,
            help="Use another password for the database then defined in settings.py",
        )
        parser.add_argument(
            "-D",
            "--dbname",
            action="store",
            dest="dbname",
            default=None,
            help='Use another database name then defined in settings.py (For PostgreSQL this defaults to "template1")',
        )
        parser.add_argument(
            "--database",
            action="store",
            dest="database",
            default=DEFAULT_DB_ALIAS,
            help='Nominates a database to synchronize. Defaults to the "default" database.',
        )
        parser.add_argument(
            "-e",
            "--exclude",
            dest="exclude",
            action="append",
            default=[],
            help="App to exclude (use multiple --exclude to exclude multiple apps).",
        )

    requires_model_validation = False

    def handle(self, *args, **options):
        database = settings.DATABASES[options.get("database", DEFAULT_DB_ALIAS)]

        if options.get("interactive"):
            confirm = input(
                """
You have requested a database reset.
This will IRREVERSIBLY DESTROY
ALL data in the database "%s".
Are you sure you want to do this?

Type 'yes' to continue, or 'no' to cancel: """
                % database["NAME"],
            )
        else:
            confirm = "yes"

        if confirm != "yes":
            print("Reset cancelled.")
            return

        self.db_handle(database, args, options)

    def db_handle(self, database, args, options):
        pass


def _get_permission_codename(action, opts):
    return "%s_%s" % (action, opts.object_name.lower())


def _get_all_permissions(opts):
    "Returns (codename, name) for all permissions in the given opts."
    perms = []
    for action in settings.DEFAULT_PERMISSIONS:
        perms.append((_get_permission_codename(action, opts), "Can %s %s" % (action, opts.verbose_name_raw)))
    return perms + list(opts.permissions)


def create_permissions(app_config, **kwargs):
    from django.contrib.auth.models import Permission
    from django.contrib.contenttypes.models import ContentType

    for klass in app_config.get_models():
        ctype = ContentType.objects.get_for_model(klass)
        for codename, name in _get_all_permissions(klass._meta):
            p, created = Permission.objects.get_or_create(
                codename=codename,
                content_type__pk=ctype.id,
                defaults={"name": name, "content_type": ctype},
            )
            if created and kwargs.get("verbosity", 0) >= 2:
                log.info("Adding permission '{}'".format(p))


signals.post_migrate.connect(create_permissions, dispatch_uid="fluo.management.create_permissions")



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

# Copyright (C) 2007 Michael Trier
# Autocomplete feature taken from django-extensions
from functools import update_wrapper

from django.contrib import admin
from django.forms.widgets import SelectMultiple
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

from ..db import models
from .nested import NestedModelAdmin, NestedStackedInline, NestedTabularInline
from .views import RelatedSearchJsonView
from .widgets import RelatedSearchSelect, RelatedSearchSelectMultiple

__all__ = [
    "ModelAdmin",
    "OrderedModelAdmin",
    "TreeOrderedModelAdmin",
    "CategoryModelAdmin",
    "StackedInline",
    "TabularInline",
    "ReadOnlyMixin",
    "ReadOnlyModelAdmin",
    "ReadOnlyInlineMixin",
    "ReadOnlyStackedInline",
    "ReadOnlyTabularInline",
]

# begin admin customization
# set this field for all models
admin.ModelAdmin.save_on_top = True
admin.options.FORMFIELD_FOR_DBFIELD_DEFAULTS[models.OrderField] = {"required": False}
for field in [
    "StringField",
    "CIStringField",
    "URLField",
    "CIURLField",
    "SlugField",
    "CISlugField",
    "EmailField",
    "CIEmailField",
]:
    if hasattr(models, field):
        admin.options.FORMFIELD_FOR_DBFIELD_DEFAULTS[getattr(models, field)] = {
            "widget": admin.widgets.AdminTextInputWidget,
        }
# end admin customization


class RelatedSearchMixin:
    related_search_fields = {}

    def get_related_search_fields(self, request):
        """
        Return a list of ForeignKey and/or ManyToMany fields which should use
        an related_search widget.
        """
        return self.related_search_fields

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if "widget" not in kwargs:
            if db_field.name in self.get_related_search_fields(request):
                kwargs["widget"] = RelatedSearchSelect(db_field, self.admin_site, using=kwargs.get("using"))
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        """
        Get a form Field for a ManyToManyField.
        """
        # If it uses an intermediary model that isn't auto created, don't show
        # a field in admin.
        if not db_field.remote_field.through._meta.auto_created:
            return None

        if "widget" not in kwargs:
            # autocomplete_fields = self.get_autocomplete_fields(request)
            # if db_field.name in autocomplete_fields:
            if db_field.name in self.get_related_search_fields(request):
                kwargs["widget"] = RelatedSearchSelectMultiple(
                    db_field,
                    self.admin_site,
                    db=kwargs.get("using"),
                )
        form_field = super().formfield_for_manytomany(db_field, request, **kwargs)
        if isinstance(form_field.widget, SelectMultiple) and not isinstance(
            form_field.widget,
            RelatedSearchSelectMultiple,
        ):
            msg = _("Hold down “Control”, or “Command” on a Mac, to select more than one.")
            help_text = form_field.help_text
            form_field.help_text = format_lazy("{} {}", help_text, msg) if help_text else msg
        return form_field

    def get_urls(self):
        from django.urls import path

        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(view)(*args, **kwargs)

            return update_wrapper(wrapper, view)

        info = self.model._meta.app_label, self.model._meta.model_name

        return [
            path(
                "related_search/<path:field_name>/",
                wrap(self.related_search_view),
                name="%s_%s_related_search" % info,
            ),
        ] + super().get_urls()

    def related_search_view(self, request, field_name):
        related_field = self.model._meta.get_field(field_name)
        related_model = related_field.remote_field.model
        model_admin = self.admin_site._registry[related_model]
        search_fields = self.get_related_search_fields(request)[field_name]
        if not search_fields and model_admin is not None:
            search_fields = model_admin.get_search_fields(request)
        return RelatedSearchJsonView.as_view(model_admin=model_admin, search_fields=search_fields)(
            request,
            related_field.name,
        )


class ModelAdmin(RelatedSearchMixin, NestedModelAdmin):
    pass


class StackedInline(RelatedSearchMixin, NestedStackedInline):
    pass


class TabularInline(RelatedSearchMixin, NestedTabularInline):
    pass


class OrderedModelAdmin(ModelAdmin):
    ordering = ["ordering"]

    def get_queryset(self, request):
        return super().get_queryset(request).order_by("ordering")


class TreeOrderedModelAdmin(OrderedModelAdmin):
    def get_queryset(self, request):
        return super().get_queryset(request).filter(parent__isnull=True)


class CategoryModelAdmin(OrderedModelAdmin):
    search_fields = ["status", "name"]
    ordering = ["ordering", "name"]
    fieldsets = [
        (_("visualization admin"), {"fields": ["ordering"], "classes": ["collapse"]}),
        (_("general admin"), {"fields": ["status", "default"]}),
        (None, {"fields": ["name"]}),
    ]


class ReadOnlyMixin:
    def get_readonly_fields(self, request, obj=None):
        return self.fields or [f.name for f in self.model._meta.fields]

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        if request.method not in ("GET", "HEAD"):
            return False
        return super().has_change_permission(request, obj)

    def has_delete_permission(self, request, obj=None):
        return False


class ReadOnlyModelAdmin(ReadOnlyMixin, admin.ModelAdmin):
    actions = None


class ReadOnlyInlineMixin(ReadOnlyMixin):
    can_delete = False
    extra = 0


class ReadOnlyStackedInline(ReadOnlyInlineMixin, StackedInline):
    pass


class ReadOnlyTabularInline(ReadOnlyInlineMixin, TabularInline):
    pass



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import logging

from django.core.exceptions import ImproperlyConfigured
from django.http import (
    HttpResponse,
    HttpResponseGone,
    HttpResponseNotAllowed,
    HttpResponsePermanentRedirect,
    HttpResponseRedirect,
)
from django.template.response import TemplateResponse
from django.utils.translation import gettext as _

log = logging.getLogger("fluo")


class View:
    METHODS = (
        "head",
        "get",
        "post",
        "put",
        "delete",
        "trace",
        "options",
        "connect",
        "trace",
        "patch",
    )
    content_type = None

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            if key in self.METHODS:
                raise TypeError(
                    "You tried to pass in the %s method name as a keyword argument to %s(). Don't do that."
                    % (key, self.__class__.__name__),
                )
            if not hasattr(self, key):
                raise TypeError(
                    "%s() received an invalid keyword %r. only accepts arguments that are already attributes of the class."  # noqa: E501
                    % (self.__class__.__name__, key),
                )
            else:
                setattr(self, key, value)

        if hasattr(self, "get") and not hasattr(self, "head"):
            self.head = self.get

        self.allowed_methods = []
        self.methods = {}
        for item in View.METHODS:
            method = getattr(self, item, None)
            if method is not None:
                self.allowed_methods.append(item.upper())
                self.methods[item] = method

    def __call__(self, request, *args, **kwargs):
        try:
            method = self.methods[request.method.lower()]
        except KeyError:
            method = self.http_method_not_allowed
        return method(request, *args, **kwargs)

    def http_method_not_allowed(self, request, *args, **kwargs):
        log.warning(
            "Method Not Allowed (%s): %s",
            request.method,
            request.path,
            extra={"status_code": 405, "request": request},
        )
        return HttpResponseNotAllowed(self.allowed_methods)

    def options(self, request, *args, **kwargs):
        """
        Handles responding to requests for the OPTIONS HTTP verb
        """
        response = HttpResponse()
        response["Allow"] = ", ".join(self.allowed_methods)
        response["Content-Length"] = 0
        return response


class TemplateView(View):
    """
    A view that renders a template.
    """

    template_name = None
    content_type = None

    def get_context_data(self, **kwargs):
        return {
            "params": kwargs,
        }

    def get_template_names(self):
        if self.template_name is not None:
            return [self.template_name]
        msg = _("%s must either define 'template_name' or override 'get_template_names()'")
        raise ImproperlyConfigured(msg % self.__class__.__name__)

    def render(self, request, template_name, context=None, content_type=None, status=None):
        return TemplateResponse(
            request=request,
            template=template_name,
            context=context,
            content_type=self.content_type if content_type is None else content_type,
            status=status,
        )

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        return self.render(request=request, template_name=self.get_template_names(), context=context)


class RedirectView(View):
    permanent = True
    url = None
    query_string = True

    def __init__(self, **kwargs):
        self.permanent = kwargs.pop("permanent", self.permanent)
        self.url = kwargs.pop("url", self.url)
        self.query_string = kwargs.pop("query_string", self.query_string)
        super().__init__(**kwargs)

    def get_redirect_url(self, request, **kwargs):
        """
        Return the URL redirect to. Keyword arguments from the
        URL pattern match generating the redirect request
        are provided as kwargs to this method.
        """
        if self.url:
            url = self.url % kwargs
            args = request.META.get("QUERY_STRING", "")
            if args and self.query_string:
                url = "%s?%s" % (url, args)
            return url
        else:
            return None

    def get(self, request, *args, **kwargs):
        url = self.get_redirect_url(request, **kwargs)
        if url:
            if self.permanent:
                return HttpResponsePermanentRedirect(url)
            else:
                return HttpResponseRedirect(url)
        else:
            log.warning(
                "Gone: %s",
                self.request.path,
                extra={"status_code": 410, "request": self.request},
            )
            return HttpResponseGone()

    def head(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def options(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from __future__ import annotations


def get_version(version=None):
    "Returns a PEP 386-compliant version number from VERSION."
    if version is None:
        from .. import VERSION as version
    else:
        assert len(version) == 5
        assert version[3] in ("alpha", "beta", "rc", "final")

    # Now build the two parts of the version number:
    # main = X.Y[.Z]
    # sub = .devN - for pre-alpha releases
    #     | {a|b|c}N - for alpha, beta and rc releases

    parts = 2 if version[2] == 0 else 3
    main = ".".join(str(x) for x in version[:parts])

    sub = ""

    if version[3] != "final":
        mapping = {"alpha": "a", "beta": "b", "rc": "c"}
        sub = mapping[version[3]] + str(version[4])

    return str(main + sub)



# Copyright (C) 2007-2022, Raffaele Salmaso <raffaele@salmaso.org>
# Copyright (c) 2015 S-Block
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from django import forms
from django.conf import settings
from django.contrib.admin import helpers
from django.contrib.admin.options import InlineModelAdmin, ModelAdmin, reverse
from django.contrib.admin.utils import unquote
from django.core.exceptions import FieldDoesNotExist, PermissionDenied
from django.db import models, transaction
from django.forms.formsets import all_valid
from django.http import Http404
from django.utils.decorators import method_decorator
from django.utils.encoding import force_str
from django.utils.html import escape
from django.utils.translation import gettext as _
from django.views.decorators.csrf import csrf_protect

__all__ = [
    "NestedModelAdmin",
    "NestedInline",
    "NestedStackedInline",
    "NestedTabularInline",
]


csrf_protect_m = method_decorator(csrf_protect)


class InlineInstancesMixin:
    def get_inline_instances(self, request, obj=None):
        inline_instances = []
        for inline_class in self.inlines:
            inline = inline_class(self.model, self.admin_site)
            if request:
                if not (
                    inline.has_add_permission(request, obj)
                    or inline.has_change_permission(request, obj)
                    or inline.has_delete_permission(request, obj)
                ):
                    continue
                if not inline.has_add_permission(request, obj):
                    inline.max_num = 0
            inline_instances.append(inline)

        return inline_instances


class NestedModelAdmin(InlineInstancesMixin, ModelAdmin):
    @property
    def media(self):
        media = super().media
        css = media._css
        css["all"] = css.get("all", []).append("fluo/admin/css/forms-nested.css")
        js = media._js + [
            f"fluo/admin/js/inlines-nested{'' if settings.DEBUG else '.min'}.js",
        ]
        return forms.Media(css=css, js=js)

    def save_formset(self, request, form, formset, change):
        """
        Given an inline formset save it to the database.
        """
        formset.save()

        for form in formset.forms:
            if hasattr(form, "nested_formsets") and form not in formset.deleted_forms:
                for nested_formset in form.nested_formsets:
                    self.save_formset(request, form, nested_formset, change)

    def save_related(self, request, form, formsets, change):
        """
        Given the ``HttpRequest``, the parent ``ModelForm`` instance, the
        list of inline formsets and a boolean value based on whether the
        parent is being added or changed, save the related objects to the
        database. Note that at this point save_form() and save_model() have
        already been called.
        """
        form.save_m2m()
        for formset in formsets:
            self.save_formset(request, form, formset, change=change)

    def add_nested_inline_formsets(self, request, inline, formset, depth=0):
        if depth > 5:
            raise Exception("Maximum nesting depth reached (5)")
        for form in formset.forms:
            nested_formsets = []
            for nested_inline in inline.get_inline_instances(request):
                InlineFormSet = nested_inline.get_formset(request, form.instance)
                prefix = "%s-%s" % (form.prefix, InlineFormSet.get_default_prefix())

                if request.method == "POST" and any(s.startswith(prefix) for s in request.POST.keys()):
                    nested_formset = InlineFormSet(
                        request.POST,
                        request.FILES,
                        instance=form.instance,
                        prefix=prefix,
                        queryset=nested_inline.get_queryset(request),
                    )
                else:
                    nested_formset = InlineFormSet(
                        instance=form.instance,
                        prefix=prefix,
                        queryset=nested_inline.get_queryset(request),
                    )
                nested_formsets.append(nested_formset)
                if nested_inline.inlines:
                    self.add_nested_inline_formsets(request, nested_inline, nested_formset, depth=depth + 1)
            form.nested_formsets = nested_formsets

    def wrap_nested_inline_formsets(self, request, inline, formset):
        media = None

        def get_media(extra_media):
            if media:
                return media + extra_media
            else:
                return extra_media

        for form in formset.forms:
            wrapped_nested_formsets = []
            for nested_inline, nested_formset in zip(inline.get_inline_instances(request), form.nested_formsets):
                if form.instance.pk:
                    instance = form.instance
                else:
                    instance = None
                fieldsets = list(nested_inline.get_fieldsets(request, instance))
                readonly = list(nested_inline.get_readonly_fields(request, instance))
                prepopulated = dict(nested_inline.get_prepopulated_fields(request, instance))
                wrapped_nested_formset = helpers.InlineAdminFormSet(
                    nested_inline,
                    nested_formset,
                    fieldsets,
                    prepopulated,
                    readonly,
                    model_admin=self,
                )
                wrapped_nested_formsets.append(wrapped_nested_formset)
                media = get_media(wrapped_nested_formset.media)
                if nested_inline.inlines:
                    media = get_media(self.wrap_nested_inline_formsets(request, nested_inline, nested_formset))
            form.nested_formsets = wrapped_nested_formsets
        return media

    def formset_has_nested_data(self, formsets):
        for formset in formsets:
            if not formset.is_bound:
                pass
            for form in formset:
                if hasattr(form, "cleaned_data") and form.cleaned_data:
                    return True
                elif hasattr(form, "nested_formsets"):
                    if self.formset_has_nested_data(form.nested_formsets):
                        return True

    def all_valid_with_nesting(self, formsets):
        "Recursively validate all nested formsets"
        if not all_valid(formsets):
            return False

        for formset in formsets:
            if not formset.is_bound:
                pass
            for form in formset:
                if hasattr(form, "nested_formsets"):
                    if not self.all_valid_with_nesting(form.nested_formsets):
                        return False

                    # TODO - find out why this breaks when extra = 1 and just adding new item with no sub items
                    if (not hasattr(form, "cleaned_data") or not form.cleaned_data) and self.formset_has_nested_data(
                        form.nested_formsets,
                    ):
                        form._errors["__all__"] = form.error_class(
                            ["Parent object must be created when creating nested inlines."],
                        )
                        return False
        return True

    @csrf_protect_m
    @transaction.atomic
    def add_view(self, request, form_url="", extra_context=None):
        "The 'add' admin view for this model."
        model = self.model
        opts = model._meta

        if not self.has_add_permission(request):
            raise PermissionDenied

        ModelForm = self.get_form(request)
        formsets = []
        inline_instances = self.get_inline_instances(request, None)
        if request.method == "POST":
            form = ModelForm(request.POST, request.FILES)
            if form.is_valid():
                new_object = self.save_form(request, form, change=False)
                form_validated = True
            else:
                form_validated = False
                new_object = self.model()
            prefixes = {}
            for FormSet, inline in self.get_formsets_with_inlines(request):
                prefix = FormSet.get_default_prefix()
                prefixes[prefix] = prefixes.get(prefix, 0) + 1
                if prefixes[prefix] != 1 or not prefix:
                    prefix = "%s-%s" % (prefix, prefixes[prefix])
                formset = FormSet(
                    data=request.POST,
                    files=request.FILES,
                    instance=new_object,
                    save_as_new="_saveasnew" in request.POST,
                    prefix=prefix,
                    queryset=inline.get_queryset(request),
                )
                formsets.append(formset)
                if inline.inlines:
                    self.add_nested_inline_formsets(request, inline, formset)
            if self.all_valid_with_nesting(formsets) and form_validated:
                self.save_model(request, new_object, form, False)
                self.save_related(request, form, formsets, False)
                add_message = self.construct_change_message(request, form, formsets, add=True)
                self.log_addition(request, new_object, add_message)
                return self.response_add(request, new_object)
        else:
            # Prepare the dict of initial data from the request.
            # We have to special-case M2Ms as a list of comma-separated PKs.
            initial = dict(request.GET.items())
            for k in initial:
                try:
                    f = opts.get_field(k)
                except FieldDoesNotExist:
                    continue
                if isinstance(f, models.ManyToManyField):
                    initial[k] = initial[k].split(",")
            form = ModelForm(initial=initial)
            prefixes = {}
            for FormSet, inline in self.get_formsets_with_inlines(request):
                prefix = FormSet.get_default_prefix()
                prefixes[prefix] = prefixes.get(prefix, 0) + 1
                if prefixes[prefix] != 1 or not prefix:
                    prefix = "%s-%s" % (prefix, prefixes[prefix])
                formset = FormSet(instance=self.model(), prefix=prefix, queryset=inline.get_queryset(request))
                formsets.append(formset)
                if hasattr(inline, "inlines") and inline.inlines:
                    self.add_nested_inline_formsets(request, inline, formset)

        adminForm = helpers.AdminForm(
            form,
            list(self.get_fieldsets(request)),
            self.get_prepopulated_fields(request),
            self.get_readonly_fields(request),
            model_admin=self,
        )
        media = self.media + adminForm.media

        inline_admin_formsets = []
        for inline, formset in zip(inline_instances, formsets):
            fieldsets = list(inline.get_fieldsets(request))
            readonly = list(inline.get_readonly_fields(request))
            prepopulated = dict(inline.get_prepopulated_fields(request))
            inline_admin_formset = helpers.InlineAdminFormSet(
                inline,
                formset,
                fieldsets,
                prepopulated,
                readonly,
                model_admin=self,
            )
            inline_admin_formsets.append(inline_admin_formset)
            media = media + inline_admin_formset.media
            if hasattr(inline, "inlines") and inline.inlines:
                extra_media = self.wrap_nested_inline_formsets(request, inline, formset)
                if extra_media:
                    media += extra_media

        context = {
            **self.admin_site.each_context(request),
            "title": _("Add %s") % force_str(opts.verbose_name),
            "adminform": adminForm,
            "is_popup": "_popup" in request.GET,
            "show_delete": False,
            "media": media,
            "inline_admin_formsets": inline_admin_formsets,
            "errors": helpers.AdminErrorList(form, formsets),
            "app_label": opts.app_label,
        }
        context.update(extra_context or {})
        return self.render_change_form(request, context, form_url=form_url, add=True)

    @csrf_protect_m
    @transaction.atomic
    def change_view(self, request, object_id, form_url="", extra_context=None):
        "The 'change' admin view for this model."
        model = self.model
        opts = model._meta

        obj = self.get_object(request, unquote(object_id))

        if not self.has_change_permission(request, obj):
            raise PermissionDenied

        if obj is None:
            raise Http404(
                _("%(name)s object with primary key %(key)r does not exist.")
                % {"name": force_str(opts.verbose_name), "key": escape(object_id)},
            )

        if request.method == "POST" and "_saveasnew" in request.POST:
            return self.add_view(
                request,
                form_url=reverse(
                    "admin:%s_%s_add" % (opts.app_label, opts.module_name),
                    current_app=self.admin_site.name,
                ),
            )

        ModelForm = self.get_form(request, obj)
        formsets = []
        inline_instances = self.get_inline_instances(request, obj)
        if request.method == "POST":
            form = ModelForm(request.POST, request.FILES, instance=obj)
            if form.is_valid():
                form_validated = True
                new_object = self.save_form(request, form, change=True)
            else:
                form_validated = False
                new_object = obj
            prefixes = {}
            for FormSet, inline in self.get_formsets_with_inlines(request, new_object):
                prefix = FormSet.get_default_prefix()
                prefixes[prefix] = prefixes.get(prefix, 0) + 1
                if prefixes[prefix] != 1 or not prefix:
                    prefix = "%s-%s" % (prefix, prefixes[prefix])
                formset = FormSet(
                    request.POST,
                    request.FILES,
                    instance=new_object,
                    prefix=prefix,
                    queryset=inline.get_queryset(request),
                )
                formsets.append(formset)
                if hasattr(inline, "inlines") and inline.inlines:
                    self.add_nested_inline_formsets(request, inline, formset)

            if self.all_valid_with_nesting(formsets) and form_validated:
                self.save_model(request, new_object, form, True)
                self.save_related(request, form, formsets, True)
                change_message = self.construct_change_message(request, form, formsets)
                self.log_change(request, new_object, change_message)
                return self.response_change(request, new_object)

        else:
            form = ModelForm(instance=obj)
            prefixes = {}
            for FormSet, inline in self.get_formsets_with_inlines(request, obj):
                prefix = FormSet.get_default_prefix()
                prefixes[prefix] = prefixes.get(prefix, 0) + 1
                if prefixes[prefix] != 1 or not prefix:
                    prefix = "%s-%s" % (prefix, prefixes[prefix])
                formset = FormSet(instance=obj, prefix=prefix, queryset=inline.get_queryset(request))
                formsets.append(formset)
                if hasattr(inline, "inlines") and inline.inlines:
                    self.add_nested_inline_formsets(request, inline, formset)

        adminForm = helpers.AdminForm(
            form,
            self.get_fieldsets(request, obj),
            self.get_prepopulated_fields(request, obj),
            self.get_readonly_fields(request, obj),
            model_admin=self,
        )
        media = self.media + adminForm.media

        inline_admin_formsets = []
        for inline, formset in zip(inline_instances, formsets):
            fieldsets = list(inline.get_fieldsets(request, obj))
            readonly = list(inline.get_readonly_fields(request, obj))
            prepopulated = dict(inline.get_prepopulated_fields(request, obj))
            inline_admin_formset = helpers.InlineAdminFormSet(
                inline,
                formset,
                fieldsets,
                prepopulated,
                readonly,
                model_admin=self,
            )
            inline_admin_formsets.append(inline_admin_formset)
            media = media + inline_admin_formset.media
            if hasattr(inline, "inlines") and inline.inlines:
                extra_media = self.wrap_nested_inline_formsets(request, inline, formset)
                if extra_media:
                    media += extra_media

        context = {
            **self.admin_site.each_context(request),
            "title": _("Change %s") % force_str(opts.verbose_name),
            "adminform": adminForm,
            "object_id": object_id,
            "original": obj,
            "is_popup": "_popup" in request.GET,
            "media": media,
            "inline_admin_formsets": inline_admin_formsets,
            "errors": helpers.AdminErrorList(form, formsets),
            "app_label": opts.app_label,
        }
        context.update(extra_context or {})
        return self.render_change_form(request, context, change=True, obj=obj, form_url=form_url)


class NestedInline(InlineInstancesMixin, InlineModelAdmin):
    inlines = []
    new_objects = []

    @property
    def media(self):
        media = super().media
        js = media._js + [
            f"fluo/admin/js/inlines-nested{'' if settings.DEBUG else '.min'}.js",
        ]
        return forms.Media(js=js)

    def get_formsets_with_inlines(self, request, obj=None):
        for inline in self.get_inline_instances(request):
            yield inline.get_formset(request, obj), inline


class NestedStackedInline(NestedInline):
    template = "admin/edit_inline/stacked-nested.html"


class NestedTabularInline(NestedInline):
    template = "admin/edit_inline/tabular-nested.html"
